@@ -51,13 +51,15 @@ __all__ = ["namemacro", "isnamemacro",
            "MacroExpander", "MacroCollector",
            "expand_macros", "find_macros"]
 
-from ast import (Name, Subscript, Tuple, Import, alias, AST, Assign, Store, Constant,
-                 Lambda, arguments, Call, copy_location, iter_fields, NodeVisitor)
+import sys
+from ast import (AST, Assign, Call, Constant, Import, Lambda, Name,
+                 NodeVisitor, Store, Subscript, Tuple, alias, arguments,
+                 copy_location, iter_fields)
 from copy import copy
 from warnings import warn_explicit
 
-from .core import BaseMacroExpander, global_postprocess, Done
-from .coreutils import ismacroimport, get_macros
+from .core import BaseMacroExpander, Done, global_postprocess
+from .coreutils import get_macros, ismacroimport
 from .unparser import unparse_with_fallbacks
 from .utils import format_macrofunction
 
@@ -99,7 +101,11 @@ def destructure_candidate(tree):
     if type(tree) is Name:
         return tree.id, []
     elif type(tree) is Subscript and type(tree.value) is Name:
-        macroargs = tree.slice.value
+        if sys.version_info >= (3, 9, 0):  # Python 3.9+: no ast.Index wrapper
+            macroargs = tree.slice
+        else:
+            macroargs = tree.slice.value
+
         if type(macroargs) is Tuple:  # [a0, a1, ...]
             macroargs = macroargs.elts
         else:  # anything that doesn't have at least one comma at the top level
@@ -141,7 +147,10 @@ class MacroExpander(BaseMacroExpander):
         macroname, macroargs = destructure_candidate(candidate)
         if self.ismacrocall(macroname, macroargs, "expr"):
             kw = {"args": macroargs}
-            tree = subscript.slice.value
+            if sys.version_info >= (3, 9, 0):  # Python 3.9+: no ast.Index wrapper
+                tree = subscript.slice
+            else:
+                tree = subscript.slice.value
             sourcecode = unparse_with_fallbacks(subscript, debug=True, color=True, expander=self)
             new_tree = self.expand("expr", subscript, macroname, tree, sourcecode=sourcecode, kw=kw)
             if new_tree is None:
@@ -388,7 +397,10 @@ class MacroCollector(NodeVisitor):
             self.visit(macroargs)
             # Don't `self.generic_visit(tree)`; that'll incorrectly detect
             # the name part as an identifier macro. Recurse only in the expr.
-            self.visit(subscript.slice.value)
+            if sys.version_info >= (3, 9, 0):  # Python 3.9+: no ast.Index wrapper
+                self.visit(subscript.slice)
+            else:
+                self.visit(subscript.slice.value)
         else:
             self.generic_visit(subscript)
 
@@ -497,7 +509,11 @@ def _make_coverage_dummy_expr(macronode):
 # --------------------------------------------------------------------------------
 
 def expand_macros(tree, bindings, *, filename):
-    """Expand `tree` with macro bindings `bindings`. Top-level entrypoint.
+    """Expand `tree` with macro bindings `bindings`. Top-level entry point.
+
+    Note that while this is a top-level entry point for the **macro** expander,
+    expanding macros is only a part of the full import algorithm. See the function
+    `mcpyrate.compiler.expand` for the 30,000ft (9,144m) view.
 
     Primarily meant to be called with `tree` the AST of a module that uses
     macros, but works with any `tree` (even inside a macro, if you need an
@@ -514,7 +530,11 @@ def expand_macros(tree, bindings, *, filename):
 
 
 def find_macros(tree, *, filename, reload=False, self_module=None, transform=True):
-    """Establish macro bindings from `tree`. Top-level entrypoint.
+    """Establish macro bindings from `tree`. Top-level entry point.
+
+    Note that while this is a top-level entry point for the **macro** expander,
+    expanding macros is only a part of the full import algorithm. See the function
+    `mcpyrate.compiler.expand` for the 30,000ft (9,144m) view.
 
     Collect bindings from each macro-import statement (`from ... import macros, ...`)
     at the top level of `tree.body`.
@@ -535,7 +555,7 @@ def find_macros(tree, *, filename, reload=False, self_module=None, transform=Tru
 
     `self_module`: str, optional, absolute dotted module name of the module being
                    expanded. Used for supporting `from __self__ import macros, ...`
-                   for multi-phase compilation (a.k.a. `with phase`).
+                   for multi-phase compilation (a.k.a. staging).
 
     `transform`: If enabled, transform each macro-import into `import ...`,
                  where `...` is the absolute module name the macros are being
@@ -553,6 +573,7 @@ def find_macros(tree, *, filename, reload=False, self_module=None, transform=Tru
 
     Return value is a dict `{macroname: function, ...}` with all collected bindings.
     """
+    stmts_to_delete = []
     bindings = {}
     for index, statement in enumerate(tree.body):
         if ismacroimport(statement):
@@ -563,11 +584,21 @@ def find_macros(tree, *, filename, reload=False, self_module=None, transform=Tru
                     # Remove self-macro-imports after establishing bindings.
                     # No need to import a module at run time; the importer lifts all
                     # the higher-phase code also into the code of the current phase.
+                    #
+                    # `statement` usually has location info, in which case we can replace the
+                    # self-macro-import with a coverage dummy node. But it might not, if we are
+                    # dealing with a dynamically generated module that's being multi-phase compiled.
+                    # In that case it's best to just delete the statement now that it's done its job.
                     dummies = _insert_coverage_dummy_stmt(None, statement, "<self-macro-import>", filename)
-                    tree.body[index] = dummies[0]
+                    if dummies is not None:  # had location info?
+                        tree.body[index] = dummies[0]
+                    else:
+                        stmts_to_delete.append(index)
                 else:
                     # Remove all names to prevent macros being used as regular run-time objects.
                     # Always use an absolute import, for the unhygienic expose API guarantee.
                     tree.body[index] = copy_location(Import(names=[alias(name=module_absname, asname=None)]),
                                                      statement)
+    for index in reversed(stmts_to_delete):
+        tree.body.pop(index)
     return bindings
