@@ -12,20 +12,24 @@ part of the public API of `mcpyrate.debug`, because it's a debugging utility.
 As a developer, see `ismultiphase` and `multiphase_expand`; the rest are
 implementation details. We provide `isdebug` and `detect_highest_phase`
 for convenient introspection.
+
+Multi-phase compilation is also known as *staging*, but the focus is usually
+slightly different. Hence we have preferred the term *phase*, from Racket,
+because Racket's phase level tower is the system that most resembles this one.
 """
 
 __all__ = ["phase", "ismultiphase", "detect_highest_phase", "isdebug", "multiphase_expand"]
 
 import ast
-from copy import copy, deepcopy
 import sys
-from types import ModuleType
+from copy import copy, deepcopy
 
-from .colorizer import setcolor, ColorScheme
+from . import compiler
+from .colorizer import ColorScheme, setcolor
 from .coreutils import ismacroimport
-from .expander import find_macros, expand_macros, destructure_candidate, namemacro, parametricmacro
-from .markers import check_no_markers_remaining
+from .expander import destructure_candidate, namemacro, parametricmacro
 from .unparser import unparse_with_fallbacks
+from .utils import getdocstring
 
 # --------------------------------------------------------------------------------
 # Private utilities.
@@ -103,7 +107,11 @@ def extract_phase(tree, *, phase=0):
         else:
             # Lifting to phase >= 1. Decrease the `n` in `with phase[n]`
             # by one, so the block gets processed again in the next phase.
-            macroarg = withphase.items[0].context_expr.slice.value
+            if sys.version_info >= (3, 9, 0):  # Python 3.9+: no ast.Index wrapper
+                macroarg = withphase.items[0].context_expr.slice
+            else:
+                macroarg = withphase.items[0].context_expr.slice.value
+
             if type(macroarg) is ast.Constant:
                 macroarg.value -= 1
             elif type(macroarg) is ast.Num:  # TODO: Python 3.8: remove ast.Num
@@ -128,7 +136,7 @@ def extract_phase(tree, *, phase=0):
 
 @parametricmacro
 def phase(tree, syntax, **kw):
-    """[syntax, block] Control multi-phase compilation.
+    """[syntax, block] Control multi-phase compilation (a.k.a. staging).
 
     Allows a module to use macros defined in that module itself.
 
@@ -304,7 +312,7 @@ def isdebug(tree):
 # --------------------------------------------------------------------------------
 # The multi-phase compiler.
 
-def multiphase_expand(tree, dexpander, *, filename, self_module, _optimize=-1):
+def multiphase_expand(tree, *, filename, self_module, dexpander=None, _optimize=-1):
     """Macro-expand an AST in multiple phases, controlled by `with phase[n]`.
 
     Primarily meant to be called with `tree` the AST of a module that
@@ -317,13 +325,15 @@ def multiphase_expand(tree, dexpander, *, filename, self_module, _optimize=-1):
     Once phase `k = 0` is reached and the code has been macro-expanded, the temporary
     entry is deleted from `sys.modules`.
 
-    `dexpander`:        The `DialectExpander` instance to use for dialect AST transforms.
-
     `filename`:         Full path to the `.py` file being compiled.
 
     `self_module`:      Absolute dotted module name of the module being compiled.
-                        Will be used to temporarily inject the temporary,
-                        higher-phase modules into `sys.modules`.
+                        Used for temporarily injecting the temporary, higher-phase
+                        modules into `sys.modules`, as well as resolving `__self__`
+                        in self-macro-imports (`from __self__ import macros, ...`).
+
+    `dexpander`:        The `DialectExpander` instance to use for dialect AST transforms.
+                        If not provided, dialect processing is skipped.
 
     `_optimize`:        Passed on to Python's built-in `compile` function, when compiling
                         the temporary higher-phase modules. Has no effect on the final result.
@@ -346,18 +356,30 @@ def multiphase_expand(tree, dexpander, *, filename, self_module, _optimize=-1):
     if debug:
         print(f"{c(CS.HEADING)}**Multi-phase compiling module {c(CS.TREEID)}'{self_module}' ({c(CS.SOURCEFILENAME)}{filename}{c(CS.TREEID)}){c()}", file=sys.stderr)
 
+    # Inject temporary module into `sys.modules`.
+    #
+    # Note that if the module already exists in `sys.modules`, its entry will be overwritten.
+    # We don't touch the parent module (if any), so that if it already refers to an old (but
+    # phase-0, fully compiled) version of this one, the reference won't be clobbered with one
+    # pointing to the temporary module.
+    #
+    # We must inject the temporary module only once, to keep the already compiled higher-phase
+    # stuff available in the module's namespace while the next phase is being compiled.
+    # (This matters when there are at least 3 phases, see `demo/let.py`.)
+    module = compiler.create_module(dotted_name=self_module, filename=filename, update_parent=False)
+
     for k in range(n, -1, -1):  # phase 0 is what a regular compile would do
         if debug:
             print(f"{c(CS.HEADING)}**AST for {c(CS.ATTENTION)}PHASE {k}{c(CS.HEADING)} of module {c(CS.TREEID)}'{self_module}' ({c(CS.SOURCEFILENAME)}{filename}{c(CS.TREEID)}){c()}", file=sys.stderr)
 
         phase_k_tree = extract_phase(tree, phase=k)
         if phase_k_tree.body:
-            # inject `__phase__ = k` for introspection
+            # inject `__phase__ = k` for introspection (at run time of the phase being compiled now)
             tgt = ast.Name(id="__phase__", ctx=ast.Store(), lineno=1, col_offset=1)
             val = ast.Constant(value=k, lineno=1, col_offset=13)
             assignment = ast.Assign(targets=[tgt], value=val, lineno=1, col_offset=1)
-            # TODO: remove ast.Str once we bump minimum language version to Python 3.8
-            if type(phase_k_tree.body[0]) is ast.Expr and type(phase_k_tree.body[0].value) in (ast.Constant, ast.Str):
+
+            if getdocstring(phase_k_tree.body):
                 docstring, *body = phase_k_tree.body
                 phase_k_tree.body = [docstring, assignment] + body
             else:  # no docstring
@@ -366,25 +388,15 @@ def multiphase_expand(tree, dexpander, *, filename, self_module, _optimize=-1):
             if debug:
                 print(unparse_with_fallbacks(phase_k_tree, debug=True, color=True), file=sys.stderr)
 
-            phase_k_tree, dialect_instances = dexpander.transform_ast(phase_k_tree)
-            module_macro_bindings = find_macros(phase_k_tree, filename=filename, self_module=self_module)
-            expansion = expand_macros(phase_k_tree, bindings=module_macro_bindings, filename=filename)
-            expansion = dexpander.postprocess_ast(expansion, dialect_instances)
-            check_no_markers_remaining(expansion, filename=filename)
-
             # Once we hit the final phase, no more temporary modules - let the import system take over.
             if k == 0:
+                expansion = compiler.singlephase_expand(phase_k_tree, filename=filename, self_module=self_module, dexpander=dexpander)
                 break
 
-            # Compile temporary module, and inject it into `sys.modules`, so we can compile the next phase.
-            #
-            # We don't bother with hifi stuff, such as attributes usually set by the importer,
-            # or even the module docstring. We essentially just want the macro functions,
-            # so that the next phase can bind to them.
-            temporary_code = compile(expansion, filename, "exec", dont_inherit=True, optimize=_optimize)
-            temporary_module = ModuleType(self_module)
-            sys.modules[self_module] = temporary_module
-            exec(temporary_code, temporary_module.__dict__)
+            # Compile the current tree, and run it in the namespace of the temporary module. This
+            # gives us the temporary higher-phase module, which allows us to compile the next phase.
+            # At intermediate phases, some definitions overwrite previous ones - this is ok.
+            compiler.run(phase_k_tree, module)
 
     # restore `sys.modules` to how it was before we began the multi-phase compile for this module
     if original_module:
