@@ -1,7 +1,7 @@
 # -*- coding: utf-8; -*-
 """Quasiquotes. Build ASTs in your macros, using syntax that mostly looks like regular code.
 
-The macro operators `q`, `u`, `n`, `a`, `s`, `h` are the primary API.
+The macro operators `q`, `u`, `n`, `a`, `s`, `t`, `h` are the primary API.
 
 The functions `capture_value` and `capture_as_macro` are public, so you can get the
 benefits of hygienic capture also in old-school macros that build ASTs manually
@@ -16,19 +16,28 @@ __all__ = ["capture_value", "capture_macro", "capture_as_macro",
            "q", "u", "n", "a", "s", "t", "h"]
 
 import ast
+import copy
 import pickle
+import sys
 
-from .core import global_bindings, Done, MacroExpansionError
+from .core import Done, MacroExpansionError, global_bindings
+from .coreutils import _mcpyrate_attr
 from .expander import MacroExpander, isnamemacro
-from .markers import ASTMarker, check_no_markers_remaining
+from .markers import ASTMarker, check_no_markers_remaining, delete_markers
 from .unparser import unparse, unparse_with_fallbacks
-from .utils import gensym, scrub_uuid, flatten, extract_bindings, NestingLevelTracker
+from .utils import (NestingLevelTracker, extract_bindings, flatten, gensym,
+                    scrub_uuid)
 
 
-def _mcpyrate_quotes_attr(attr):
-    """Create an AST that, when compiled and run, looks up `mcpyrate.quotes.attr`."""
-    mcpyrate_quotes_module = ast.Attribute(value=ast.Name(id="mcpyrate"), attr="quotes")
-    return ast.Attribute(value=mcpyrate_quotes_module, attr=attr)
+def _mcpyrate_quotes_attr(attr, *, force_import=False):
+    """Create an AST that, when compiled and run, looks up `mcpyrate.quotes.attr`.
+
+    If `force_import` is `True`, use the builtin `__import__` function to
+    first import the `mcpyrate.quotes` module. This is useful for e.g.
+    hygienically unquoted values, whose eventual use site might not import
+    any `mcpyrate` modules.
+    """
+    return _mcpyrate_attr(f"quotes.{attr}", force_import=force_import)
 
 
 class QuasiquoteMarker(ASTMarker):
@@ -43,6 +52,19 @@ class SpliceNodes(QuasiquoteMarker):
     """
     pass
 
+class QuasiquoteSearchDone(Done, QuasiquoteMarker):
+    """Marker used by nested quasiquotes to tell the expander a subtree is already done.
+
+    This inherits, but is separate, from the usual `Done`, because:
+
+      1. We need to tell the expander that is processing the nested quasiquotes
+         to stop expanding an invocation that has already been considered.
+
+      2. We need to be able to eliminate these (and only these) before
+         generating the final quoted output.
+    """
+    pass
+
 # --------------------------------------------------------------------------------
 # Unquote commands for `astify`. Each type corresponds to an unquote macro.
 
@@ -145,7 +167,7 @@ def ast_literal(tree, syntax):
     where splicing into the surrounding context is needed.
     """
     if syntax not in ("expr", "block"):
-        raise TypeError(f"expected `syntax` either 'expr' or 'block', got {type(syntax)} with value {repr(syntax)}")
+        raise ValueError(f"expected `syntax` either 'expr' or 'block', got {repr(syntax)}")
 
     if syntax == "expr":
         if isinstance(tree, ast.AST):
@@ -264,7 +286,7 @@ def capture_value(value, name):
     # and serialization.
     #
     frozen_value = pickle.dumps(value)
-    return ast.Call(_mcpyrate_quotes_attr("lookup_value"),
+    return ast.Call(_mcpyrate_quotes_attr("lookup_value", force_import=True),
                     [ast.Tuple(elts=[ast.Constant(value=name),
                                      ast.Constant(value=frozen_value)])],
                     [])
@@ -444,7 +466,8 @@ def astify(x, expander=None):  # like `macropy`'s `ast_repr`
 
         # We must support at least the `Done` AST marker, so that things like
         # coverage dummy nodes and expanded name macros can be astified.
-        elif isinstance(x, Done):
+        # (Note we support only exactly `Done`, not arbitrary descendants.)
+        elif T is Done:
             fields = [ast.keyword(a, recurse(b)) for a, b in ast.iter_fields(x)]
             node = ast.Call(_mcpyrate_quotes_attr("Done"),
                             [],
@@ -522,12 +545,12 @@ def unastify(tree):
     our_module_globals = globals()
     def lookup_thing(dotted_name):
         if not dotted_name.startswith("mcpyrate.quotes"):
-            raise NotImplementedError
+            raise NotImplementedError(f"Don't know how to look up {repr(dotted_name)}")
         path = dotted_name.split(".")
         if not all(component.isidentifier() for component in path):
-            raise NotImplementedError
+            raise NotImplementedError(f"Dotted name {repr(dotted_name)} contains at least one non-identifier component")
         if len(path) < 3:
-            raise NotImplementedError
+            raise NotImplementedError(f"Dotted name {repr(dotted_name)} has fewer than two dots (expected 'mcpyrate.quotes.something')")
         name_of_thing = path[2]
         thing = our_module_globals[name_of_thing]
         if len(path) > 3:
@@ -560,6 +583,12 @@ def unastify(tree):
     elif T is ast.Call:
         dotted_name = unparse(tree.func)
 
+        # Drop the run-time part of `q`, if present. This is added by `q` itself,
+        # not `astify`, but `unastify` is usually applied to the output of `q`.
+        if dotted_name == "mcpyrate.quotes.splice_ast_literals":  # `q[]`
+            body = tree.args[0]
+            return unastify(body)
+
         # Even though the unquote operators compile into calls, `unastify`
         # must not apply their run-time parts, because it's running in the
         # wrong context. Those only work properly at run time, and they
@@ -568,7 +597,7 @@ def unastify(tree):
         #
         # So we undo what `astify` did, converting the unquote calls back into
         # the corresponding AST markers.
-        if dotted_name == "mcpyrate.quotes.astify":  # `u[]`
+        elif dotted_name == "mcpyrate.quotes.astify":  # `u[]`
             body = tree.args[0]
             return Unquote(body)
         elif dotted_name == "mcpyrate.quotes.lift_sourcecode":  # `n[]`
@@ -587,7 +616,7 @@ def unastify(tree):
             body, name = tree.args[0], tree.args[1].value
             return Capture(body, name)
         elif dotted_name == "mcpyrate.quotes.lookup_macro":  # `h[]` (macro)
-            # `capture_macro` is gone and done by the time we get here.
+            # `capture_macro` is done and gone by the time we get here.
             # `astify` has generated an `ast.Call` to `lookup_macro`.
             #
             # To make the this work properly even across process boundaries,
@@ -650,15 +679,45 @@ def _expand_quasiquotes(tree, expander):
     bindings = extract_bindings(expander.bindings, q, u, n, a, s, t, h)
     return MacroExpander(bindings, expander.filename).visit(tree)
 
+# TODO: maybe make the rest of this a method of `MacroExpander`, and only wrap with `QuasiquoteSearchDone` here?
+def _replace_tree_in_macro_invocation(invocation, newtree):
+    """Helper function for handling nested quasiquotes.
 
-def q(tree, *, syntax, expander, **kw):
+    Output a new invocation of the same macro, but wrapped in `QuasiquoteSearchDone`,
+    and with the `tree` inside replaced by `newtree`.
+
+    `expr` and `block` modes are supported; this is autodetected from `invocation`.
+    """
+    new_invocation = copy.copy(invocation)
+    if type(new_invocation) is ast.Subscript:
+        if sys.version_info >= (3, 9, 0):  # Python 3.9+: no ast.Index wrapper
+            new_invocation.slice = newtree
+        else:
+            new_invocation.slice = copy.copy(invocation.slice)
+            new_invocation.slice.value = newtree
+    elif type(new_invocation) is ast.With:
+        new_invocation.body = newtree
+    else:
+        raise NotImplementedError
+    return QuasiquoteSearchDone(body=new_invocation)
+
+
+def q(tree, *, syntax, expander, invocation, **kw):
     """[syntax, expr/block] quasiquote. Lift code into its AST representation."""
     if syntax not in ("expr", "block"):
         raise SyntaxError("`q` is an expr and block macro only")
     with _quotelevel.changed_by(+1):
-        tree = _expand_quasiquotes(tree, expander)  # expand any inner quotes and unquotes first
+        tree = _expand_quasiquotes(tree, expander)  # expand any unquotes corresponding to this level first
+        if _quotelevel.value > 1:  # nested inside an outer quote?
+            # TODO: Implications when in block mode and not the only context manager in the `with`?
+            # TODO: Probably doesn't work in that case. Document that `q`, when used,
+            # TODO: should be the only ctxmgr in that particular `with`.
+            return _replace_tree_in_macro_invocation(invocation, tree)
+
+        tree = delete_markers(tree, cls=QuasiquoteSearchDone)
+
         tree = astify(tree, expander)  # Magic part of `q`. Supply `expander` for `h[macro]` detection.
-        # `astify` should compile the unquote commands away, and `SpliceNodes`
+        # `astify` should compile the unquote command markers away, and `SpliceNodes`
         # markers only spring into existence when the run-time part of `a` runs
         # (for communication with the run-time part of the surrounding `q`).
         # So at this point, there should be no quasiquote markers in `tree`.
@@ -696,7 +755,7 @@ def q(tree, *, syntax, expander, **kw):
         return tree
 
 
-def u(tree, *, syntax, expander, **kw):
+def u(tree, *, syntax, expander, invocation, **kw):
     """[syntax, expr] unquote. Splice a simple value into a quasiquote.
 
     The value is lifted into an AST that re-constructs that value.
@@ -707,10 +766,12 @@ def u(tree, *, syntax, expander, **kw):
         raise SyntaxError("`u` encountered while quotelevel < 1")
     with _quotelevel.changed_by(-1):
         tree = expander.visit_recursively(tree)
+        if _quotelevel.value > 0:
+            return _replace_tree_in_macro_invocation(invocation, tree)
         return Unquote(tree)
 
 
-def n(tree, *, syntax, expander, **kw):
+def n(tree, *, syntax, expander, invocation, **kw):
     """[syntax, expr] name-unquote. Parse a string, as Python source code, into an AST.
 
     With `n[]`, you can e.g. compute a name (e.g. by `mcpyrate.gensym`) for a
@@ -734,10 +795,12 @@ def n(tree, *, syntax, expander, **kw):
         raise SyntaxError("`n` encountered while quotelevel < 1")
     with _quotelevel.changed_by(-1):
         tree = expander.visit_recursively(tree)
+        if _quotelevel.value > 0:
+            return _replace_tree_in_macro_invocation(invocation, tree)
         return LiftSourcecode(tree, expander.filename)
 
 
-def a(tree, *, syntax, expander, **kw):
+def a(tree, *, syntax, expander, invocation, **kw):
     """[syntax, expr/block] ast-unquote. Splice an AST into a quasiquote.
 
     **Expression mode**::
@@ -789,6 +852,9 @@ def a(tree, *, syntax, expander, **kw):
 
     with _quotelevel.changed_by(-1):
         tree = expander.visit_recursively(tree)
+        if _quotelevel.value > 0:
+            # TODO: implications for block mode?
+            return _replace_tree_in_macro_invocation(invocation, tree)
 
         if syntax == "expr":
             return ASTLiteral(tree, syntax)
@@ -796,7 +862,7 @@ def a(tree, *, syntax, expander, **kw):
         assert syntax == "block"
         # Block mode: strip `Expr` wrappers.
         #
-        # When `a` expands, the elements of the list `tree` are `Expr` nodes
+        # When `with a` expands, the elements of the list `tree` are `Expr` nodes
         # containing expressions. Typically each expression is just a `Name`
         # node, or in general, any expression that at run time evaluates to
         # a statement AST node, or to an iterable of statement AST nodes.
@@ -816,7 +882,7 @@ def a(tree, *, syntax, expander, **kw):
         return ASTLiteral(ast.List(elts=out), syntax)
 
 
-def s(tree, *, syntax, expander, **kw):
+def s(tree, *, syntax, expander, invocation, **kw):
     """[syntax, expr] ast-list-unquote. Splice an iterable of ASTs, as an `ast.List`, into a quasiquote."""
     if syntax != "expr":
         raise SyntaxError("`s` is an expr macro only")
@@ -824,10 +890,12 @@ def s(tree, *, syntax, expander, **kw):
         raise SyntaxError("`s` encountered while quotelevel < 1")
     with _quotelevel.changed_by(-1):
         tree = expander.visit_recursively(tree)
+        if _quotelevel.value > 0:
+            return _replace_tree_in_macro_invocation(invocation, tree)
         return ASTList(tree)
 
 
-def t(tree, *, syntax, expander, **kw):
+def t(tree, *, syntax, expander, invocation, **kw):
     """[syntax, expr] ast-tuple-unquote. Splice an iterable of ASTs, as an `ast.Tuple`, into a quasiquote."""
     if syntax != "expr":
         raise SyntaxError("`t` is an expr macro only")
@@ -835,10 +903,12 @@ def t(tree, *, syntax, expander, **kw):
         raise SyntaxError("`t` encountered while quotelevel < 1")
     with _quotelevel.changed_by(-1):
         tree = expander.visit_recursively(tree)
+        if _quotelevel.value > 0:
+            return _replace_tree_in_macro_invocation(invocation, tree)
         return ASTTuple(tree)
 
 
-def h(tree, *, syntax, expander, **kw):
+def h(tree, *, syntax, expander, invocation, **kw):
     """[syntax, expr] hygienic-unquote. Splice any value, from the macro definition site, into a quasiquote.
 
     Supports also values that have no meaningful `repr`. The value is captured
@@ -864,6 +934,10 @@ def h(tree, *, syntax, expander, **kw):
     with _quotelevel.changed_by(-1):
         name = unparse(tree)
 
+        # TODO: This logic never does anything - any correctly placed `h[]`
+        # is always inside a `q`, which recurses with an expander that
+        # only knows about the quasiquote macros.
+
         # Expand macros in the unquoted expression. The only case we need to
         # look out for is a `@namemacro` if we have a `h[macroname]`. We're
         # only capturing it, so don't expand it just yet.
@@ -875,4 +949,7 @@ def h(tree, *, syntax, expander, **kw):
         if expand:
             tree = expander.visit_recursively(tree)
 
+        if _quotelevel.value > 0:
+            return _replace_tree_in_macro_invocation(invocation, tree)
+
         return Capture(tree, name)
