@@ -5,13 +5,13 @@ __all__ = ["UnparserError", "unparse", "unparse_with_fallbacks"]
 
 import ast
 import builtins
-from contextlib import contextmanager
 import io
 import sys
+from contextlib import contextmanager
 
-from .astdumper import dump  # fallback
-from .colorizer import setcolor, colorize, ColorScheme
 from . import markers
+from .astdumper import dump  # fallback
+from .colorizer import ColorScheme, colorize, setcolor
 
 # Large float and imaginary literals get turned into infinities in the AST.
 # We unparse those infinities to INFSTR.
@@ -53,6 +53,17 @@ class Unparser:
                  debug=False, color=False, expander=None):
         """Print the source for `tree` to `file`.
 
+        `tree`:  an AST, or a `list` of ASTs.
+
+                 If a `list`, each element is processed separately,
+                 preserving ordering. This can be used e.g. to look at
+                 the source code corresponding to a statement suite in
+                 an AST.
+
+        `file`:  where to send the unparsed source code. Must have `write`
+                 and `flush` methods. Typically something like `sys.stdout`,
+                 or an `io.StringIO` instance.
+
         `debug`: bool, print invisible nodes (`Module`, `Expr`).
 
                  For statement nodes, print also line numbers (`lineno`
@@ -73,7 +84,11 @@ class Unparser:
         self.expander = expander
         self.f = file
         self._indent = 0
-        self.dispatch(tree)
+        if isinstance(tree, list):  # statement suite (for unparsing those directly)
+            for elt in tree:
+                self.dispatch(elt)
+        else:
+            self.dispatch(tree)
         print("", file=self.f)
         self.f.flush()
 
@@ -227,10 +242,21 @@ class Unparser:
         if print_mode == "expr":
             self.write(")")
 
-    def _Module(self, t):
+    # top level nodes
+    def _Module(self, t):  # ast.parse(..., mode="exec")
+        self.toplevelnode(t)
+
+    def _Interactive(self, t):  # ast.parse(..., mode="single")
+        self.toplevelnode(t)
+
+    def _Expression(self, t):  # ast.parse(..., mode="eval")
+        self.toplevelnode(t)
+
+    def toplevelnode(self, t):
         # TODO: Python 3.8 type_ignores. Since we don't store the source text, maybe ignore that?
         if self.debug:
-            self.fill(self.maybe_colorize("$Module", ColorScheme.INVISIBLENODE),
+            label = f"${t.__class__.__name__}"
+            self.fill(self.maybe_colorize(label, ColorScheme.INVISIBLENODE),
                       lineno_node=t)
             self.enter()
             for stmt in t.body:
@@ -369,12 +395,12 @@ class Unparser:
         for ex in t.handlers:
             self.dispatch(ex)
         if t.orelse:
-            self.fill(self.maybe_colorize_python_keyword("else"))
+            self.fill(self.maybe_colorize_python_keyword("else"), lineno_node=t)
             self.enter()
             self.dispatch(t.orelse)
             self.leave()
         if t.finalbody:
-            self.fill(self.maybe_colorize_python_keyword("finally"))
+            self.fill(self.maybe_colorize_python_keyword("finally"), lineno_node=t)
             self.enter()
             self.dispatch(t.finalbody)
             self.leave()
@@ -478,7 +504,7 @@ class Unparser:
         self.dispatch(t.body)
         self.leave()
         if t.orelse:
-            self.fill(self.maybe_colorize_python_keyword("else"))
+            self.fill(self.maybe_colorize_python_keyword("else"), lineno_node=t)
             self.enter()
             self.dispatch(t.orelse)
             self.leave()
@@ -494,14 +520,14 @@ class Unparser:
         while (t.orelse and len(t.orelse) == 1 and
                isinstance(t.orelse[0], ast.If)):
             t = t.orelse[0]
-            self.fill(self.maybe_colorize_python_keyword("elif "))
+            self.fill(self.maybe_colorize_python_keyword("elif "), lineno_node=t)
             self.dispatch(t.test)
             self.enter()
             self.dispatch(t.body)
             self.leave()
         # final else
         if t.orelse:
-            self.fill(self.maybe_colorize_python_keyword("else"))
+            self.fill(self.maybe_colorize_python_keyword("else"), lineno_node=t)
             self.enter()
             self.dispatch(t.orelse)
             self.leave()
@@ -513,7 +539,7 @@ class Unparser:
         self.dispatch(t.body)
         self.leave()
         if t.orelse:
-            self.fill(self.maybe_colorize_python_keyword("else"))
+            self.fill(self.maybe_colorize_python_keyword("else"), lineno_node=t)
             self.enter()
             self.dispatch(t.orelse)
             self.leave()
@@ -800,7 +826,7 @@ class Unparser:
     def _Ellipsis(self, t):  # up to Python 3.7
         self.write("...")
 
-    def _Index(self, t):
+    def _Index(self, t):  # up to Python 3.8; the Index wrapper is gone in Python 3.9
         self.dispatch(t.value)
 
     def _Slice(self, t):
@@ -813,7 +839,7 @@ class Unparser:
             self.write(":")
             self.dispatch(t.step)
 
-    def _ExtSlice(self, t):
+    def _ExtSlice(self, t):  # up to Python 3.8; Python 3.9 uses a Tuple instead
         interleave(lambda: self.write(", "), self.dispatch, t.dims)
 
     # argument
